#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <unordered_map>
#include <vector>

#include "graph.h"
#include "graphUtils.h"
#include "traversals.h"


using namespace std;

int nV;
int nE;


void readGraph(istream& in, Graph& g)
{
	in >> nV >> nE;
	for (int i = 0; i < nV; ++i)
		add_vertex(g);
	for (int i = 0; i < nE; ++i)
	{
		int src, dest, cap;
		in >> src >> dest >> cap;
		Edge e = add_edge(src, dest, g).first;
		g[e].capacity = cap;
		g[e].flow = 0;
	}
}

ostream& operator<< (ostream& out, const vector<Vertex>& v)
{
	out << "[";
	for (unsigned i = 0; i < v.size()-1; ++i)
	{
		if (i > 0)
			out << ", ";
		out << '(' << v[i] << ',' << v[i+1] << ')';
	}
	out << "]";
	return out;
}

int smallestUnusedCapacityAlong(const vector<Vertex>& path, const Graph& g)
{
	int minExcess = INT_MAX;
	for (unsigned i = 1; i < path.size(); ++i)
	{
		Vertex v = path[i-1];
		Vertex w = path[i];
		Edge e = edge(v, w, g).first;
		minExcess = min(minExcess, g[e].capacity - abs(g[e].flow));
	}
	return minExcess;
}

void addToFlows (int delta, const vector<Vertex>& path, Graph& g)
{
	for (unsigned i = 1; i < path.size(); ++i)
	{
		Vertex v = path[i-1];
		Vertex w = path[i];
		Edge e = edge(v, w, g).first;
		if (v < w)
			g[e].flow += delta;
		else
			g[e].flow -= delta;
	}

}

int totalFlow(const Graph& g)
{
	auto outgoing = out_edges(0, g);
	int sum = 0;
	for (auto e = outgoing.first; e != outgoing.second; ++e)
	{
		sum += abs(g[*e].flow);
	}
	return sum;
}


vector<Vertex> findAugmentingPath (Graph& g)
{
	for (Vertex v = 0; v < num_vertices(g); ++v)
		g[v].visited = false;
	queue<Vertex> q;
	q.push(0);
	g[0].visited = true;
	int goal = num_vertices(g)-1;
	while (!q.empty() && !g[goal].visited)
	{
		Vertex v = q.front();
		q.pop();
		auto outgoing = out_edges(v, g);
		for (auto e = outgoing.first; e != outgoing.second; ++e)
		{
			Vertex w = target(*e, g);
			cerr << "Looking at " << w << endl;
			if (!g[w].visited)
			{
				if (v < w)
				{
					if (g[*e].flow >= g[*e].capacity)
						continue;
				}
				else // v > w
				{
					if (-g[*e].flow >= g[*e].capacity)
						continue;
				}
				q.push(w);
				g[w].visited = true;
				g[w].cameFrom = v;
			}
		}
	}
	vector<Vertex> path;
	if (g[goal].visited)
	{
		path.push_back(goal);
		Vertex w = goal;
		while (w != 0)
		{
			w = g[w].cameFrom;
			path.push_back(w);
		}
		reverse(path.begin(), path.end());
	}
	if (path.size() > 0)
		cerr << "Augmenting? " << path << endl;
	return path;
}

int maxFlow (Graph& g)
{
	while (true) {
	   // find an "augmenting path" from 0 to |V|-1 for which each edge has unused capacity
	   vector<Vertex> path = findAugmentingPath(g);
	   if (path.size() == 0)
		   break;
	   // set delta to smallest unused capacity along that path
	   int delta = smallestUnusedCapacityAlong(path, g);
	   cerr << "delta is " << delta << endl;
	   // add delta to the flow of each edge along that path
	   addToFlows (delta, path, g);
	}
	// answer is sum of the flows coming in to the warehouse
	return totalFlow(g);
}

void solve (Graph& g)
{
	int flow = maxFlow (g);
	cout << "Max flow is " << flow << endl;
}


int main (int argc, char** argv)
{
	Graph citymap;
	if (argc > 1)
	{
		ifstream in (argv[1]);
		readGraph (in, citymap);
	}
	else
	{
		readGraph (cin, citymap);
	}
	solve (citymap);
	return 0;

}
