#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <unordered_map>
#include <vector>

#include "graph.h"
#include "graphUtils.h"
#include "traversals.h"


using namespace std;

int nV;
int nE;


void readGraph(istream& in, Graph& g)
{
	in >> nV >> nE;
	for (int i = 0; i < nV; ++i)
		add_vertex(g);
	for (int i = 0; i < nE; ++i)
	{
		int src, dest, cap;
		in >> src >> dest >> cap;
		Edge e = add_edge(src, dest, g).first;
		g[e].capacity = cap;
		g[e].flow = 0;
	}
}

ostream& operator<< (ostream& out, const vector<Vertex>& v)
{
	out << "[";
	for (unsigned i = 0; i < v.size()-1; ++i)
	{
		if (i > 0)
			out << ", ";
		cout << '(' << v[i] << ',' << v[i+1] << ')';
	}
	out << "]";
	return out;
}

int smallestUnusedCapacityAlong(const vector<Vertex>& path, const Graph& g)
{
	int minCap = INT_MAX;
	for (unsigned i = 0; i < path.size()-1; ++i)
	{
		Vertex v = path[i];
		Vertex w = path[i+1];
		Edge e = edge(v, w, g).first;
		int cap = g[e].capacity - abs(g[e].flow);
		minCap = min(minCap, cap);
	}
	return minCap;
}

void addToFlows (int delta, const vector<Vertex>& path, Graph& g)
{
	for (unsigned i = 0; i < path.size()-1; ++i)
	{
		Vertex v = path[i];
		Vertex w = path[i+1];
		Edge e = edge(v, w, g).first;
		if (v < w)
			g[e].flow += delta;
		else
			g[e].flow -= delta;
	}
}

int totalFlow(const Graph& g)
{
	int total = 0;
	auto outEdges = out_edges((Vertex)0, g);
	for (auto e = outEdges.first; e != outEdges.second; ++e)
	{
		total += g[*e].flow;
	}
	return total;
}


vector<Vertex> findAugmentingPath (Graph& g)
{
	auto allVertices = vertices(g);
	for (auto v = allVertices.first; v != allVertices.second; ++v)
	{
		g[*v].visited = false;
		g[*v].cameFrom = nV;
	}

	queue<Vertex, list<Vertex> > q;
	q.push(0);
	g[0].visited = true;

	while (!q.empty())
	{
		Vertex v = q.front();
		q.pop();
		auto outgoing = out_edges(v,g);
		for (auto e = outgoing.first; e != outgoing.second; ++e)
		{
			Vertex w = target(*e, g);
			if (v < w)
			{
				// positive flow
				int excess = g[*e].capacity - g[*e].flow;
				if (excess <= 0) continue;
			}
			else
			{
				// negative flow
				int excess = g[*e].capacity + g[*e].flow;
				if (excess <= 0) continue;				
			}
			if (!g[w].visited)
			{
				q.push (w);
				g[w].visited = true;
				g[w].cameFrom = v;
				if (w == (Vertex)nV-1) break;
			}
		}
	}
	vector<Vertex> path;
	if (g[nV-1].visited)
	{
		path.push_back((Vertex)nV-1);
		Vertex v = nV -1;
		while (v != 0)
		{
			v = g[v].cameFrom;
			path.push_back(v);
		}
		reverse (path.begin(), path.end());
		cerr << "aug path. " << path << endl;
	}
	return path;
}

int maxFlow (Graph& g)
{
	while (true)
	{
   		vector<Vertex> path = findAugmentingPath(g);
		if (path.size() == 0) break;
		int delta = smallestUnusedCapacityAlong(path, g);
		cerr << "delta: " << delta << endl;
		addToFlows(delta, path, g);
	}
	return totalFlow(g);
}

vector<Edge> minCut(const Graph& g)
{
	vector<Edge> results;
	for (Vertex v = 0; v < num_vertices(g); ++v)
	{
		auto outgoing = out_edges(v, g);
		for (auto e = outgoing.first; e != outgoing.second; ++e)
		{
			Vertex w = target(*e, g);
			if (v < w && g[v].visited != g[w].visited)
				results.push_back(*e);
		}
	}
	return results;
}

void solve (Graph& g)
{
	int flow = maxFlow (g);
	cout << "Max flow is " << flow << endl;
	vector<Edge> cuts = minCut(g);
	for (const Edge& e: cuts)
	{
		cout << "Cut (" << source(e, g) << ',' << target(e,g) << ')' << endl;
	}
}


int main (int argc, char** argv)
{
	Graph citymap;
	if (argc > 1)
	{
		ifstream in (argv[1]);
		readGraph (in, citymap);
	}
	else
	{
		readGraph (cin, citymap);
	}
	solve (citymap);
	return 0;

}
